import importlib
import os
from pathlib import Path
import sys
from typing import Any, NewType
import warnings
import copy

import google.protobuf.internal.api_implementation

with warnings.catch_warnings():
    # Suppress the eth-typing package`s warnings related to some new networks
    warnings.filterwarnings("ignore", "Network .* does not have a valid ChainId. eth-typing should be "
                                      "updated with the latest networks.", UserWarning)

from snet.sdk.metadata_provider.ipfs_metadata_provider import IPFSMetadataProvider
from snet.sdk.payment_strategies.default_payment_strategy import DefaultPaymentStrategy
from snet.cli.commands.sdk_command import SDKCommand
from snet.cli.commands.commands import BlockchainCommand
from snet.cli.config import Config
from snet.cli.utils.utils import bytes32_to_str, type_converter

google.protobuf.internal.api_implementation.Type = lambda: 'python'

from google.protobuf import symbol_database as _symbol_database

_sym_db = _symbol_database.Default()
_sym_db.RegisterMessage = lambda x: None


from urllib.parse import urljoin


import web3
from rfc3986 import urlparse
import ipfshttpclient

from snet.sdk.service_client import ServiceClient
from snet.sdk.account import Account
from snet.sdk.mpe.mpe_contract import MPEContract

from snet.contracts import get_contract_object

from snet.cli.metadata.service import mpe_service_metadata_from_json
from snet.cli.utils.ipfs_utils import bytesuri_to_hash, get_from_ipfs_and_checkhash
from snet.cli.utils.utils import find_file_by_keyword

ModuleName = NewType('ModuleName', str)
ServiceStub = NewType('ServiceStub', Any)


class Arguments:
    def __init__(self, org_id=None, service_id=None):
        self.org_id = org_id
        self.service_id = service_id
        self.language = "python"
        self.protodir = Path("~").expanduser().joinpath(".snet")


class SnetSDK:
    """Base Snet SDK"""

    def __init__(self, sdk_config, metadata_provider=None):
        self._sdk_config = sdk_config
        self._metadata_provider = metadata_provider

        # Instantiate Ethereum client
        eth_rpc_endpoint = self._sdk_config.get("eth_rpc_endpoint",
                                                "https://mainnet.infura.io/v3/e7732e1f679e461b9bb4da5653ac3fc2")
        eth_rpc_request_kwargs = self._sdk_config.get("eth_rpc_request_kwargs")

        provider = web3.HTTPProvider(endpoint_uri=eth_rpc_endpoint, request_kwargs=eth_rpc_request_kwargs)

        self.web3 = web3.Web3(provider)

        # Get MPE contract address from config if specified; mostly for local testing
        _mpe_contract_address = self._sdk_config.get("mpe_contract_address", None)
        if _mpe_contract_address is None:
            self.mpe_contract = MPEContract(self.web3)
        else:
            self.mpe_contract = MPEContract(self.web3, _mpe_contract_address)

        # Instantiate IPFS client
        ipfs_endpoint = self._sdk_config.get("default_ipfs_endpoint", "/dns/ipfs.singularitynet.io/tcp/80/")
        self.ipfs_client = ipfshttpclient.connect(ipfs_endpoint)

        # Get Registry contract address from config if specified; mostly for local testing
        _registry_contract_address = self._sdk_config.get("registry_contract_address", None)
        if _registry_contract_address is None:
            self.registry_contract = get_contract_object(self.web3, "Registry")
        else:
            self.registry_contract = get_contract_object(self.web3, "Registry", _registry_contract_address)

        self.account = Account(self.web3, sdk_config, self.mpe_contract)

        global_config = Config(sdk_config=sdk_config)
        self.setup_config(global_config)

    def setup_config(self, config: Config) -> None:
        out_f = sys.stdout
        network = self._sdk_config.get("network", None)
        identity_name = self._sdk_config.get("identity_name", None)
        # Checking for an empty network
        if network and config["session"]["network"] != network:
            config.set_session_network(network, out_f)
        if identity_name:
            self.set_session_identity(identity_name, config, out_f)
        elif len(config.get_all_identities_names()) > 0:
            if "identity" not in config["session"] or config["session"]["identity"] == "":
                raise Exception("identity_name is not passed or selected")

    def set_session_identity(self, identity_name: str, config: Config, out_f):
        if identity_name not in config.get_all_identities_names():
            identity = config.setup_identity()
            config.add_identity(identity_name, identity, out_f)
            config.set_session_identity(identity_name, out_f)
        elif config["session"]["identity"] != identity_name:
            config.set_session_identity(identity_name, out_f)

    def create_service_client(self, org_id: str, service_id: str, group_name=None,
                              payment_channel_management_strategy=None,
                              free_call_auth_token_bin=None,
                              free_call_token_expiry_block=None,
                              options=None,
                              concurrent_calls=1):

        # Create and instance of the Config object, so we can create an instance of SDKCommand
        sdk_config_object = Config(sdk_config=self._sdk_config)
        sdk = SDKCommand(sdk_config_object, args=Arguments(org_id, service_id))

        # Download the proto file and generate stubs if needed
        force_update = self._sdk_config.get('force_update', False)
        if force_update:
            sdk.generate_client_library()
        else:
            path_to_pb_files = self.get_path_to_pb_files(org_id, service_id)
            pb_2_file_name = find_file_by_keyword(path_to_pb_files, keyword="pb2.py")
            pb_2_grpc_file_name = find_file_by_keyword(path_to_pb_files, keyword="pb2_grpc.py")
            if not pb_2_file_name or not pb_2_grpc_file_name:
                sdk.generate_client_library()
        
        if payment_channel_management_strategy is None:
            payment_channel_management_strategy = DefaultPaymentStrategy(concurrent_calls)

        if options is None:
            options = dict()
        options['free_call_auth_token-bin'] = bytes.fromhex(free_call_auth_token_bin) if\
            free_call_token_expiry_block else ""
        options['free-call-token-expiry-block'] = free_call_token_expiry_block if\
            free_call_token_expiry_block else 0
        options['email'] = self._sdk_config.get("email", "")
        options['concurrency'] = self._sdk_config.get("concurrency", True)

        if self._metadata_provider is None:
            self._metadata_provider = IPFSMetadataProvider(self.ipfs_client, self.registry_contract)

        service_metadata = self._metadata_provider.enhance_service_metadata(org_id, service_id)
        group = self._get_service_group_details(service_metadata, group_name)
        strategy = payment_channel_management_strategy
        
        service_stub = self.get_service_stub(org_id, service_id)
        
        pb2_module = self.get_module_by_keyword(org_id, service_id, keyword="pb2.py")
        
        service_client = ServiceClient(org_id, service_id, service_metadata, group, service_stub, strategy,
                                       options, self.mpe_contract, self.account, self.web3, pb2_module)
        return service_client

    def get_service_stub(self, org_id: str, service_id: str) -> ServiceStub:
        path_to_pb_files = self.get_path_to_pb_files(org_id, service_id)
        module_name = self.get_module_by_keyword(org_id, service_id, keyword="pb2_grpc.py")
        package_path = os.path.dirname(path_to_pb_files)
        sys.path.append(package_path)
        try:
            grpc_file = importlib.import_module(module_name)
            properties_and_methods_of_grpc_file = dir(grpc_file)
            class_stub = [elem for elem in properties_and_methods_of_grpc_file if 'Stub' in elem][0]
            service_stub = getattr(grpc_file, class_stub)
            return ServiceStub(service_stub)
        except Exception as e:
            raise Exception(f"Error importing module: {e}")

    def get_path_to_pb_files(self, org_id: str, service_id: str) -> str:
        client_libraries_base_dir_path = Path("~").expanduser().joinpath(".snet")
        path_to_pb_files = f"{client_libraries_base_dir_path}/{org_id}/{service_id}/python/"
        return path_to_pb_files
    
    def get_module_by_keyword(self, org_id: str, service_id: str, keyword: str) -> ModuleName:
        path_to_pb_files = self.get_path_to_pb_files(org_id, service_id)
        file_name = find_file_by_keyword(path_to_pb_files, keyword)
        module_name = os.path.splitext(file_name)[0]
        return ModuleName(module_name)

    def get_service_metadata(self, org_id, service_id):
        (found, registration_id, metadata_uri) = self.registry_contract.functions.getServiceRegistrationById(
            bytes(org_id, "utf-8"),
            bytes(service_id, "utf-8")
        ).call()

        if found is not True:
            raise Exception('No service "{}" found in organization "{}"'.format(service_id, org_id))

        metadata_hash = bytesuri_to_hash(metadata_uri)
        metadata_json = get_from_ipfs_and_checkhash(self.ipfs_client, metadata_hash)
        metadata = mpe_service_metadata_from_json(metadata_json)
        return metadata

    def _get_first_group(self, service_metadata):
        return service_metadata['groups'][0]

    def _get_group_by_group_name(self, service_metadata, group_name):
        for group in service_metadata['groups']:
            if group['group_name'] == group_name:
                return group
        return {}

    def _get_service_group_details(self, service_metadata, group_name):
        if len(service_metadata['groups']) == 0:
            raise Exception("No Groups found for given service, Please add group to the service")

        if group_name is None:
            return self._get_first_group(service_metadata)

        return self._get_group_by_group_name(service_metadata, group_name)

    def get_organization_list(self) -> list:
        global_config = Config(sdk_config=self._sdk_config)
        blockchain_command = BlockchainCommand(config=global_config, args=Arguments())
        org_list = blockchain_command.call_contract_command(
            "Registry", "listOrganizations", [])
        organization_list = []
        for idx, org_id in enumerate(org_list):
            organization_list.append(bytes32_to_str(org_id))
        return organization_list

    def get_services_list(self, org_id: str) -> list:
        global_config = Config(sdk_config=self._sdk_config)
        blockchain_command = BlockchainCommand(config=global_config, args=Arguments())
        (found, org_service_list) = blockchain_command.call_contract_command("Registry",
                                                                             "listServicesForOrganization",
                                                               [type_converter("bytes32")(org_id)])
        if not found:
            raise Exception(f"Organization with id={org_id} doesn't exist!")
        org_service_list = list(map(bytes32_to_str, org_service_list))
        return org_service_list




import importlib
from urllib.parse import urlparse

import grpc
import web3
from snet.cli.resources.root_certificate import certificate
from snet.cli.utils.utils import RESOURCES_PATH, add_to_path

from snet.sdk.payment_strategies.payment_staregy import PaymentStrategy


class FreeCallPaymentStrategy(PaymentStrategy):

    def is_free_call_available(self, service_client):
        try:
            org_id, service_id, group_id, daemon_endpoint = service_client.get_service_details()
            email, token_for_free_call, token_expiry_date_block = service_client.get_free_call_config()

            if not token_for_free_call:
                return False

            signature, current_block_number = self.generate_signature(service_client)
            with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
                state_service_pb2 = importlib.import_module("state_service_pb2")

            with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
                state_service_pb2_grpc = importlib.import_module("state_service_pb2_grpc")

            request = state_service_pb2.FreeCallStateRequest()
            request.user_id = email
            request.token_for_free_call = token_for_free_call
            request.token_expiry_date_block = token_expiry_date_block
            request.signature = signature
            request.current_block = current_block_number

            endpoint_object = urlparse(daemon_endpoint)
            if endpoint_object.port is not None:
                channel_endpoint = endpoint_object.hostname + ":" + str(endpoint_object.port)
            else:
                channel_endpoint = endpoint_object.hostname

            if endpoint_object.scheme == "http":
                channel = grpc.insecure_channel(channel_endpoint)
            elif endpoint_object.scheme == "https":
                channel = grpc.secure_channel(channel_endpoint, grpc.ssl_channel_credentials(root_certificates=certificate))
            else:
                raise ValueError('Unsupported scheme in service metadata ("{}")'.format(endpoint_object.scheme))

            stub = state_service_pb2_grpc.FreeCallStateServiceStub(channel)
            response = stub.GetFreeCallsAvailable(request)
            if response.free_calls_available > 0:
                return True
            return False
        except Exception as e:
            return False

    def get_payment_metadata(self, service_client):
        email, token_for_free_call, token_expiry_date_block = service_client.get_free_call_config()
        signature, current_block_number = self.generate_signature(service_client)
        metadata = [("snet-free-call-auth-token-bin", token_for_free_call),
                    ("snet-free-call-token-expiry-block", str(token_expiry_date_block)),
                    ("snet-payment-type", "free-call"),
                    ("snet-free-call-user-id", email),
                    ("snet-current-block-number", str(current_block_number)),
                    ("snet-payment-channel-signature-bin", signature)]

        return metadata

    def select_channel(self, service_client):
        pass

    def generate_signature(self, service_client):
        org_id, service_id, group_id, daemon_endpoint = service_client.get_service_details()
        email, token_for_free_call, token_expiry_date_block = service_client.get_free_call_config()

        if token_expiry_date_block == 0 or len(email) == 0 or len(token_for_free_call) == 0:
            raise Exception(
                "You are using default 'FreeCallPaymentStrategy' to use this strategy you need to pass "
                "'free_call_auth_token-bin','email','free-call-token-expiry-block' in config")

        current_block_number = service_client.get_current_block_number()

        message = web3.Web3.solidity_keccak(
            ["string", "string", "string", "string", "string", "uint256", "bytes32"],
            ["__prefix_free_trial", email, org_id, service_id, group_id, current_block_number,
             token_for_free_call]
        )
        return service_client.generate_signature(message), current_block_number



import json
import web3

from snet.cli.metadata.service import mpe_service_metadata_from_json
from snet.cli.utils.ipfs_utils import bytesuri_to_hash, get_from_ipfs_and_checkhash


class IPFSMetadataProvider(object):

    def __init__(self, ipfs_client, registry_contract):
        self.registry_contract = registry_contract
        self._ipfs_client = ipfs_client

    def fetch_org_metadata(self, org_id):
        org = web3.Web3.to_bytes(text=org_id).ljust(32, b"\0")
        found, id, metadata_uri, owner, members, service_ids = self.registry_contract.functions.getOrganizationById(
            org
        ).call()
        if found is not True:
            raise Exception('Organization with org ID "{}" not found '.format(org_id))

        metadata_hash = bytesuri_to_hash(metadata_uri)
        metadata_json = get_from_ipfs_and_checkhash(self._ipfs_client, metadata_hash)
        org_metadata = json.loads(metadata_json)
        return org_metadata

    def fetch_service_metadata(self, org_id, service_id):
        org = web3.Web3.to_bytes(text=org_id).ljust(32, b"\0")
        service = web3.Web3.to_bytes(text=service_id).ljust(32, b"\0")

        found, registration_id, metadata_uri = self.registry_contract.functions.getServiceRegistrationById(
            org, service).call()

        if found is not True:
            raise Exception('No service "{}" found in organization "{}"'.format(service_id, org_id))

        metadata_hash = bytesuri_to_hash(metadata_uri)
        metadata_json = get_from_ipfs_and_checkhash(self._ipfs_client, metadata_hash)
        metadata = mpe_service_metadata_from_json(metadata_json)
        return metadata

    def enhance_service_metadata(self, org_id, service_id):
        service_metadata = self.fetch_service_metadata(org_id, service_id)
        org_metadata = self.fetch_org_metadata(org_id)

        org_group_map = {}
        for group in org_metadata['groups']:
            org_group_map[group['group_name']] = group

        for group in service_metadata.m['groups']:
            # merge service group with org_group
            group['payment'] = org_group_map[group['group_name']]['payment']

        return service_metadata



import base64
import collections
import importlib
import re
import os
from pathlib import Path

import grpc
import web3
from eth_account.messages import defunct_hash_message
from rfc3986 import urlparse
from snet.cli.resources.root_certificate import certificate
from snet.cli.utils.utils import RESOURCES_PATH, add_to_path, find_file_by_keyword

import snet.sdk.generic_client_interceptor as generic_client_interceptor
from snet.sdk.mpe.payment_channel_provider import PaymentChannelProvider


class _ClientCallDetails(
    collections.namedtuple(
        '_ClientCallDetails',
        ('method', 'timeout', 'metadata', 'credentials')),
    grpc.ClientCallDetails):
    pass


class ServiceClient:
    def __init__(self, org_id, service_id, service_metadata, group, service_stub, payment_strategy,
                 options, mpe_contract, account, sdk_web3, pb2_module):
        self.org_id = org_id
        self.service_id = service_id
        self.options = options
        self.group = group
        self.service_metadata = service_metadata

        self.payment_strategy = payment_strategy
        self.expiry_threshold = self.group["payment"]["payment_expiration_threshold"]
        self.__base_grpc_channel = self._get_grpc_channel()
        self.grpc_channel = grpc.intercept_channel(self.__base_grpc_channel,
                                                   generic_client_interceptor.create(self._intercept_call))
        self.payment_channel_provider = PaymentChannelProvider(sdk_web3,
                                                               self._generate_payment_channel_state_service_client(),
                                                               mpe_contract)
        self.service = self._generate_grpc_stub(service_stub)
        self.pb2_module = importlib.import_module(pb2_module) if isinstance(pb2_module, str) else pb2_module
        self.payment_channels = []
        self.last_read_block = 0
        self.account = account
        self.sdk_web3 = sdk_web3
        self.mpe_address = mpe_contract.contract.address

    def call_rpc(self, rpc_name: str, message_class: str, **kwargs):
        rpc_method = getattr(self.service, rpc_name)
        request = getattr(self.pb2_module, message_class)(**kwargs)
        return rpc_method(request)

    def _get_payment_expiration_threshold_for_group(self):
        pass

    def _generate_grpc_stub(self, service_stub):
        grpc_channel = self.__base_grpc_channel
        disable_blockchain_operations = self.options.get("disable_blockchain_operations", False)
        if disable_blockchain_operations is False:
            grpc_channel = self.grpc_channel
        stub_instance = service_stub(grpc_channel)
        return stub_instance

    def get_grpc_base_channel(self):
        return self.__base_grpc_channel

    def _get_grpc_channel(self):
        endpoint = self.options.get("endpoint", None)
        if endpoint is None:
            endpoint = self.service_metadata.get_all_endpoints_for_group(self.group["group_name"])[0]
        endpoint_object = urlparse(endpoint)
        if endpoint_object.port is not None:
            channel_endpoint = endpoint_object.hostname + ":" + str(endpoint_object.port)
        else:
            channel_endpoint = endpoint_object.hostname

        if endpoint_object.scheme == "http":
            return grpc.insecure_channel(channel_endpoint)
        elif endpoint_object.scheme == "https":
            return grpc.secure_channel(channel_endpoint,
                                       grpc.ssl_channel_credentials(root_certificates=certificate))
        else:
            raise ValueError('Unsupported scheme in service metadata ("{}")'.format(endpoint_object.scheme))

    def _get_service_call_metadata(self):
        metadata = self.payment_strategy.get_payment_metadata(self)
        return metadata

    def _intercept_call(self, client_call_details, request_iterator, request_streaming,
                        response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.extend(self._get_service_call_metadata())
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    def _filter_existing_channels_from_new_payment_channels(self, new_payment_channels):
        new_channels_to_be_added = []

        # need to change this logic ,use maps to manage channels so that we can easily navigate it
        for new_payment_channel in new_payment_channels:
            existing_channel = False
            for existing_payment_channel in self.payment_channels:
                if new_payment_channel.channel_id == existing_payment_channel.channel_id:
                    existing_channel = True
                    break

            if not existing_channel:
                new_channels_to_be_added.append(new_payment_channel)

        return new_channels_to_be_added

    def load_open_channels(self):
        current_block_number = self.sdk_web3.eth.block_number
        payment_address = self.group["payment"]["payment_address"]
        group_id = base64.b64decode(str(self.group["group_id"]))
        new_payment_channels = self.payment_channel_provider.get_past_open_channels(self.account, payment_address,
                                                                                    group_id, self.last_read_block)
        self.payment_channels = self.payment_channels + self._filter_existing_channels_from_new_payment_channels(
            new_payment_channels)
        self.last_read_block = current_block_number
        return self.payment_channels

    def get_current_block_number(self):
        return self.sdk_web3.eth.block_number

    def update_channel_states(self):
        for channel in self.payment_channels:
            channel.sync_state()
        return self.payment_channels

    def default_channel_expiration(self):
        current_block_number = self.sdk_web3.eth.get_block("latest").number
        return current_block_number + self.expiry_threshold

    def _generate_payment_channel_state_service_client(self):
        grpc_channel = self.__base_grpc_channel
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            state_service_pb2_grpc = importlib.import_module("state_service_pb2_grpc")
        return state_service_pb2_grpc.PaymentChannelStateServiceStub(grpc_channel)

    def open_channel(self, amount, expiration):
        payment_address = self.group["payment"]["payment_address"]
        group_id = base64.b64decode(str(self.group["group_id"]))
        return self.payment_channel_provider.open_channel(self.account, amount, expiration, payment_address,
                                                          group_id)

    def deposit_and_open_channel(self, amount, expiration):
        payment_address = self.group["payment"]["payment_address"]
        group_id = base64.b64decode(str(self.group["group_id"]))
        return self.payment_channel_provider.deposit_and_open_channel(self.account, amount, expiration,
                                                                      payment_address, group_id)

    def get_price(self):
        return self.group["pricing"][0]["price_in_cogs"]

    def generate_signature(self, message):
        signature = bytes(self.sdk_web3.eth.account.signHash(defunct_hash_message(message),
                                                             self.account.signer_private_key).signature)

        return signature

    def generate_training_signature(self, text: str, address, block_number):
        message = web3.Web3.solidity_keccak(
            ["string", "address", "uint256"],
            [text, address, block_number]
        )
        return self.sdk_web3.eth.account.signHash(defunct_hash_message(message),
                                                  self.account.signer_private_key).signature

    def get_free_call_config(self):
        return self.options['email'], self.options['free_call_auth_token-bin'], self.options[
            'free-call-token-expiry-block']

    def get_service_details(self):
        return self.org_id, self.service_id, self.group["group_id"], \
            self.service_metadata.get_all_endpoints_for_group(self.group["group_name"])[0]

    def get_concurrency_flag(self):
        return self.options.get('concurrency', True)

    def get_concurrency_token_and_channel(self):
        return self.payment_strategy.get_concurrency_token_and_channel(self)

    def set_concurrency_token_and_channel(self, token, channel):
        self.payment_strategy.concurrency_token = token
        self.payment_strategy.channel = channel

    def get_path_to_pb_files(self, org_id: str, service_id: str) -> str:
        client_libraries_base_dir_path = Path("~").expanduser().joinpath(".snet")
        path_to_pb_files = f"{client_libraries_base_dir_path}/{org_id}/{service_id}/python/"
        return path_to_pb_files

    def get_services_and_messages_info(self):
        # Get proto file filepath and open it
        path_to_pb_files = self.get_path_to_pb_files(self.org_id, self.service_id)
        proto_file_name = find_file_by_keyword(path_to_pb_files, keyword=".proto")
        proto_filepath = os.path.join(path_to_pb_files, proto_file_name)
        with open(proto_filepath, 'r') as file:
            proto_content = file.read()
        # Regular expression patterns to match services, methods, messages, and fields
        service_pattern = re.compile(r'service\s+(\w+)\s*{')
        rpc_pattern = re.compile(r'rpc\s+(\w+)\s*\((\w+)\)\s+returns\s+\((\w+)\)')
        message_pattern = re.compile(r'message\s+(\w+)\s*{')
        field_pattern = re.compile(r'\s*(\w+)\s+(\w+)\s*=\s*\d+\s*;')

        services = {}
        messages = {}
        current_service = None
        current_message = None

        for line in proto_content.splitlines():
            # Match a service definition
            service_match = service_pattern.search(line)
            if service_match:
                current_service = service_match.group(1)
                services[current_service] = []
                continue

            # Match an RPC method inside a service
            if current_service:
                rpc_match = rpc_pattern.search(line)
                if rpc_match:
                    method_name = rpc_match.group(1)
                    input_type = rpc_match.group(2)
                    output_type = rpc_match.group(3)
                    services[current_service].append((method_name, input_type, output_type))

            # Match a message definition
            message_match = message_pattern.search(line)
            if message_match:
                current_message = message_match.group(1)
                messages[current_message] = []
                continue

            # Match a field inside a message
            if current_message:
                field_match = field_pattern.search(line)
                if field_match:
                    field_type = field_match.group(1)
                    field_name = field_match.group(2)
                    messages[current_message].append((field_type, field_name))

        return services, messages

    def get_services_and_messages_info_as_pretty_string(self):
        services, messages = self.get_services_and_messages_info()

        string_output = ""
        # Prettify the parsed services and their methods
        for service, methods in services.items():
            string_output += f"Service: {service}\n"
            for method_name, input_type, output_type in methods:
                string_output += f"  Method: {method_name}, Input: {input_type}, Output: {output_type}\n"

        # Prettify the messages and their fields
        for message, fields in messages.items():
            string_output += f"Message: {message}\n"
            for field_type, field_name in fields:
                string_output += f"  Field: {field_type} {field_name}\n"
        return string_output



from snet.sdk.concurrency_manager import ConcurrencyManager
from snet.sdk.payment_strategies.freecall_payment_strategy import FreeCallPaymentStrategy
from snet.sdk.payment_strategies.paidcall_payment_strategy import PaidCallPaymentStrategy
from snet.sdk.payment_strategies.prepaid_payment_strategy import PrePaidPaymentStrategy
from snet.sdk.payment_strategies.payment_staregy import PaymentStrategy


class DefaultPaymentStrategy(PaymentStrategy):

    def __init__(self, concurrent_calls=1):
        self.concurrent_calls = concurrent_calls
        self.concurrencyManager = ConcurrencyManager(concurrent_calls)
        self.channel = None

    def set_concurrency_token(self, token):
        self.concurrencyManager.__token = token

    def set_channel(self, channel):
        self.channel = channel

    def get_payment_metadata(self, service_client):
        free_call_payment_strategy = FreeCallPaymentStrategy()

        if free_call_payment_strategy.is_free_call_available(service_client):
            metadata = free_call_payment_strategy.get_payment_metadata(service_client)
        else:
            if service_client.get_concurrency_flag():
                payment_strategy = PrePaidPaymentStrategy(self.concurrencyManager)
                metadata = payment_strategy.get_payment_metadata(service_client, self.channel)
            else:
                payment_strategy = PaidCallPaymentStrategy()
                metadata = payment_strategy.get_payment_metadata(service_client)

        return metadata

    def get_price(self, service_client):
        pass

    def get_concurrency_token_and_channel(self, service_client):
        payment_strategy = PrePaidPaymentStrategy(self.concurrencyManager)
        return payment_strategy.get_concurrency_token_and_channel(service_client)



class MetadataProvider(object):

    def fetch_org_metadata(self,org_id):
        pass


    def fetch_service_metadata(self,org_id,service_id):
        pass

    def enhance_service_group_details(self, service_metadata, org_metadata):
        pass




import os
from setuptools import find_namespace_packages, setup

from version import __version__

PACKAGE_NAME = 'snet.sdk'


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


with open("./requirements.txt") as f:
    requirements_str = f.read()
requirements = requirements_str.split("\n")


setup(
    name=PACKAGE_NAME,
    version=__version__,
    packages=find_namespace_packages(include=['snet.*']),
    namespace_packages=['snet'],
    url='https://github.com/singnet/snet-sdk-python',
    author='SingularityNET Foundation',
    author_email='info@singularitynet.io',
    description='SingularityNET Python SDK',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    python_requires='>=3.10',
    install_requires=requirements,
    include_package_data=True
)



import json

from snet.cli.utils.utils import get_address_from_private, normalize_private_key
from snet.contracts import get_contract_object

DEFAULT_GAS = 300000
TRANSACTION_TIMEOUT = 500


class TransactionError(Exception):
    """Raised when an Ethereum transaction receipt has a status of 0. Can provide a custom message. Optionally includes receipt"""

    def __init__(self, message, receipt=None):
        super().__init__(message)
        self.message = message
        self.receipt = receipt

    def __str__(self):
        return self.message


class Account:
    def __init__(self, w3, config, mpe_contract):
        self.config = config
        self.web3 = w3
        self.mpe_contract = mpe_contract
        _token_contract_address = self.config.get("token_contract_address", None)
        if _token_contract_address is None:
            self.token_contract = get_contract_object(
                self.web3, "SingularityNetToken")
        else:
            self.token_contract = get_contract_object(
                self.web3, "SingularityNetToken", _token_contract_address)

        private_key = config.get("private_key", None)
        signer_private_key = config.get("signer_private_key", None)
        if private_key is not None:
            self.private_key = normalize_private_key(config["private_key"])
        if signer_private_key is not None:
            self.signer_private_key = normalize_private_key(
                config["signer_private_key"])
        else:
            self.signer_private_key = self.private_key
        self.address = get_address_from_private(self.private_key)
        self.signer_address = get_address_from_private(self.signer_private_key)
        self.nonce = 0

    def _get_nonce(self):
        nonce = self.web3.eth.get_transaction_count(self.address)
        if self.nonce >= nonce:
            nonce = self.nonce + 1
        self.nonce = nonce
        return nonce

    def _get_gas_price(self):
        gas_price = self.web3.eth.gas_price
        if gas_price <= 15000000000:
            gas_price += gas_price * 1 / 3
        elif gas_price > 15000000000 and gas_price <= 50000000000:
            gas_price += gas_price * 1 / 5
        elif gas_price > 50000000000 and gas_price <= 150000000000:
            gas_price += 7000000000
        elif gas_price > 150000000000:
            gas_price += gas_price * 1 / 10
        return int(gas_price)

    def _send_signed_transaction(self, contract_fn, *args):
        transaction = contract_fn(*args).build_transaction({
            "chainId": int(self.web3.net.version),
            "gas": DEFAULT_GAS,
            "gasPrice": self._get_gas_price(),
            "nonce": self._get_nonce()
        })
        signed_txn = self.web3.eth.account.sign_transaction(
            transaction, private_key=self.private_key)
        return self.web3.to_hex(self.web3.eth.send_raw_transaction(signed_txn.rawTransaction))

    def send_transaction(self, contract_fn, *args):
        txn_hash = self._send_signed_transaction(contract_fn, *args)
        return self.web3.eth.wait_for_transaction_receipt(txn_hash, TRANSACTION_TIMEOUT)

    def _parse_receipt(self, receipt, event, encoder=json.JSONEncoder):
        if receipt.status == 0:
            raise TransactionError("Transaction failed", receipt)
        else:
            return json.dumps(dict(event().processReceipt(receipt)[0]["args"]), cls=encoder)

    def escrow_balance(self):
        return self.mpe_contract.balance(self.address)

    def deposit_to_escrow_account(self, amount_in_cogs):
        already_approved = self.allowance()
        if amount_in_cogs > already_approved:
            self.approve_transfer(amount_in_cogs)
        return self.mpe_contract.deposit(self, amount_in_cogs)

    def approve_transfer(self, amount_in_cogs):
        return self.send_transaction(self.token_contract.functions.approve, self.mpe_contract.contract.address,
                                     amount_in_cogs)

    def allowance(self):
        return self.token_contract.functions.allowance(self.address, self.mpe_contract.contract.address).call()



import enum
import importlib
from urllib.parse import urlparse

import grpc
import web3

from snet.cli.resources.root_certificate import certificate
from snet.cli.utils.utils import add_to_path, RESOURCES_PATH


# for local debug
# from snet.snet_cli.resources.proto import training_pb2_grpc
# from snet.snet_cli.resources.proto import training_pb2


# from daemon code
class ModelMethodMessage(enum.Enum):
    CreateModel = "__CreateModel"
    GetModelStatus = "__GetModelStatus"
    UpdateModelAccess = "__UpdateModelAccess"
    GetAllModels = "__UpdateModelAccess"
    DeleteModel = "__GetModelStatus"


class TrainingModel:

    def __init__(self):
        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            self.training_pb2 = importlib.import_module("training_pb2")

        with add_to_path(str(RESOURCES_PATH.joinpath("proto"))):
            self.training_pb2_grpc = importlib.import_module("training_pb2_grpc")

    def _invoke_model(self, service_client, msg: ModelMethodMessage):
        org_id, service_id, group_id, daemon_endpoint = service_client.get_service_details()

        endpoint_object = urlparse(daemon_endpoint)
        if endpoint_object.port is not None:
            channel_endpoint = endpoint_object.hostname + ":" + str(endpoint_object.port)
        else:
            channel_endpoint = endpoint_object.hostname

        if endpoint_object.scheme == "http":
            print("creating http channel: ", channel_endpoint)
            channel = grpc.insecure_channel(channel_endpoint)
        elif endpoint_object.scheme == "https":
            channel = grpc.secure_channel(channel_endpoint,
                                          grpc.ssl_channel_credentials(root_certificates=certificate))
        else:
            raise ValueError('Unsupported scheme in service metadata ("{}")'.format(endpoint_object.scheme))

        current_block_number = service_client.get_current_block_number()
        signature = service_client.generate_training_signature(msg.value, web3.Web3.to_checksum_address(
            service_client.account.address), current_block_number)
        auth_req = self.training_pb2.AuthorizationDetails(signature=bytes(signature),
                                                          current_block=current_block_number,
                                                          signer_address=service_client.account.address,
                                                          message=msg.value)
        return auth_req, channel

    # params from AI-service: status, model_id
    # params pass to daemon: grpc_service_name, grpc_method_name, address_list,
    # description, model_name, training_data_link, is_public_accessible
    def create_model(self, service_client, grpc_method_name: str, model_name: str,
                     description: str = '',
                     training_data_link: str = '', grpc_service_name='service',
                     is_publicly_accessible=False, address_list: list[str] = None):
        if address_list is None:
            address_list = []
        try:
            auth_req, channel = self._invoke_model(service_client, ModelMethodMessage.CreateModel)
            model_details = self.training_pb2.ModelDetails(grpc_method_name=grpc_method_name, description=description,
                                                           training_data_link=training_data_link,
                                                           grpc_service_name=grpc_service_name,
                                                           model_name=model_name, address_list=address_list,
                                                           is_publicly_accessible=is_publicly_accessible)
            stub = self.training_pb2_grpc.ModelStub(channel)
            response = stub.create_model(
                self.training_pb2.CreateModelRequest(authorization=auth_req, model_details=model_details))
            return response
        except Exception as e:
            print("Exception: ", e)
            return e

    # params from AI-service: status
    # params to daemon: grpc_service_name, grpc_method_name, model_id
    def get_model_status(self, service_client, model_id: str, grpc_method_name: str, grpc_service_name='service'):
        try:
            auth_req, channel = self._invoke_model(service_client, ModelMethodMessage.GetModelStatus)
            model_details = self.training_pb2.ModelDetails(grpc_method_name=grpc_method_name,
                                                           grpc_service_name=grpc_service_name, model_id=str(model_id))
            stub = self.training_pb2_grpc.ModelStub(channel)
            response = stub.get_model_status(
                self.training_pb2.ModelDetailsRequest(authorization=auth_req, model_details=model_details))
            return response
        except Exception as e:
            print("Exception: ", e)
            return e

    # params from AI-service: status
    # params to daemon: grpc_service_name, grpc_method_name, model_id
    def delete_model(self, service_client, model_id: str, grpc_method_name: str,
                     grpc_service_name='service'):
        try:
            auth_req, channel = self._invoke_model(service_client, ModelMethodMessage.DeleteModel)
            model_details = self.training_pb2.ModelDetails(grpc_method_name=grpc_method_name,
                                                           grpc_service_name=grpc_service_name, model_id=str(model_id))
            stub = self.training_pb2_grpc.ModelStub(channel)
            response = stub.delete_model(
                self.training_pb2.UpdateModelRequest(authorization=auth_req, update_model_details=model_details))
            return response
        except Exception as e:
            print("Exception: ", e)
            return e

    # params from AI-service: None
    # params to daemon: grpc_service_name, grpc_method_name, model_id, address_list, is_public, model_name, desc
    # all params required
    def update_model_access(self, service_client, model_id: str, grpc_method_name: str,
                            model_name: str, is_public: bool,
                            description: str, grpc_service_name: str = 'service', address_list: list[str] = None):
        try:
            auth_req, channel = self._invoke_model(service_client, ModelMethodMessage.UpdateModelAccess)
            model_details = self.training_pb2.ModelDetails(grpc_method_name=grpc_method_name, description=description,
                                                           grpc_service_name=grpc_service_name,
                                                           address_list=address_list,
                                                           is_publicly_accessible=is_public, model_name=model_name,
                                                           model_id=str(model_id))
            stub = self.training_pb2_grpc.ModelStub(channel)
            response = stub.update_model_access(
                self.training_pb2.UpdateModelRequest(authorization=auth_req, update_model_details=model_details))
            return response
        except Exception as e:
            print("Exception: ", e)
            return e

    # params from AI-service: None
    # params to daemon: grpc_service_name, grpc_method_name
    def get_all_models(self, service_client, grpc_method_name: str, grpc_service_name='service'):
        try:
            auth_req, channel = self._invoke_model(service_client, ModelMethodMessage.GetAllModels)
            stub = self.training_pb2_grpc.ModelStub(channel)
            response = stub.get_all_models(
                self.training_pb2.AccessibleModelsRequest(authorization=auth_req, grpc_service_name=grpc_service_name,
                                                          grpc_method_name=grpc_method_name))
            return response
        except Exception as e:
            print("Exception: ", e)
            return e


from web3._utils.events import get_event_data
from eth_abi.codec import ABICodec

from snet.sdk.mpe.payment_channel import PaymentChannel
from snet.contracts import get_contract_deployment_block


BLOCKS_PER_BATCH = 5000


class PaymentChannelProvider(object):
    def __init__(self, w3, payment_channel_state_service_client, mpe_contract):
        self.web3 = w3

        self.mpe_contract = mpe_contract
        self.event_topics = [self.web3.keccak(
            text="ChannelOpen(uint256,uint256,address,address,address,bytes32,uint256,uint256)").hex()]
        self.deployment_block = get_contract_deployment_block(self.web3, "MultiPartyEscrow")
        self.payment_channel_state_service_client = payment_channel_state_service_client

    def get_past_open_channels(self, account, payment_address, group_id, starting_block_number=0, to_block_number=None):
        if to_block_number is None:
            to_block_number = self.web3.eth.block_number

        if starting_block_number == 0:
            starting_block_number = self.deployment_block

        codec: ABICodec = self.web3.codec

        logs = []
        from_block = starting_block_number
        while from_block <= to_block_number:
            to_block = min(from_block + BLOCKS_PER_BATCH, to_block_number)
            logs = logs + self.web3.eth.get_logs({"fromBlock": from_block, "toBlock": to_block,
                                                 "address": self.mpe_contract.contract.address,
                                                 "topics": self.event_topics})
            from_block = to_block + 1

        event_abi = self.mpe_contract.contract._find_matching_event_abi(event_name="ChannelOpen")
        channels_opened = list(filter(
            lambda
                channel: (channel.sender == account.address or channel.signer == account.signer_address) and channel.recipient ==
                         payment_address and channel.groupId == group_id,

            [get_event_data(codec, event_abi, l)["args"] for l in logs]
        ))
        return list(map(lambda channel: PaymentChannel(channel["channelId"], self.web3, account,
                                                       self.payment_channel_state_service_client, self.mpe_contract),
                        channels_opened))

    def open_channel(self, account, amount, expiration, payment_address, group_id):

        receipt = self.mpe_contract.open_channel(account, payment_address, group_id, amount, expiration)
        return self._get_newly_opened_channel(receipt, account, payment_address, group_id)

    def deposit_and_open_channel(self, account, amount, expiration, payment_address, group_id):
        receipt = self.mpe_contract.deposit_and_open_channel(account, payment_address, group_id, amount,
                                                             expiration)
        return self._get_newly_opened_channel(receipt, account, payment_address, group_id)

    def _get_newly_opened_channel(self, receipt,account, payment_address, group_id):
        open_channels = self.get_past_open_channels(account, payment_address, group_id, receipt["blockNumber"],
                                                    receipt["blockNumber"])
        if len(open_channels) == 0:
            raise Exception(f"Error while opening channel, please check transaction {receipt.transactionHash.hex()} ")
        return open_channels[0]
